@@ -3,11 +3,8 @@
 # See file LICENSE for full license details.
 
 import logging
-from wirepas_messaging.gateway.api import (
-    GatewayResultCode,
-    ScratchpadStatus,
-    ScratchpadType,
-)
+import wirepas_mesh_messaging as wmm
+
 from gi.repository import GLib
 from .return_code import ReturnCode
 
@@ -78,14 +75,14 @@ class Sink:
                 self.logger.error("Cannot send message err=%s", res)
                 return ReturnCode.error_from_dbus_return_code(res)
         except GLib.Error as e:
-            self.logger.exception("Fail to send message: %s", e.message)
-            return ReturnCode.error_from_dbus_exception(e.message)
+            self.logger.exception("Fail to send message: %s", str(e))
+            return ReturnCode.error_from_dbus_exception(str(e))
         except OverflowError:
             # It may happens as protobuf has bigger container value
             self.logger.error("Invalid range value")
-            return GatewayResultCode.GW_RES_INVALID_PARAM
+            return wmm.GatewayResultCode.GW_RES_INVALID_PARAM
 
-        return GatewayResultCode.GW_RES_OK
+        return wmm.GatewayResultCode.GW_RES_OK
 
     def _on_stack_started(self, sender, object, iface, signal, params):
         # pylint: disable=unused-argument
@@ -125,7 +122,7 @@ class Sink:
         try:
             config["started"] = (self.proxy.StackStatus & 0x01) == 0
         except GLib.Error as e:
-            error = ReturnCode.error_from_dbus_exception(e.message)
+            error = ReturnCode.error_from_dbus_exception(str(e))
             self.logger.exception("Cannot get Stack state: %s", error)
             return None
 
@@ -183,24 +180,24 @@ class Sink:
                 value,
                 key,
                 self.sink_id,
-                e.message,
+                str(e),
             )
-            return ReturnCode.error_from_dbus_exception(e.message)
+            return ReturnCode.error_from_dbus_exception(str(e))
         except OverflowError:
             # It may happens as protobuf has bigger container value
             self.logger.error(
                 "Invalid range value for param %s with value %s", key, value
             )
-            return GatewayResultCode.GW_RES_INVALID_PARAM
+            return wmm.GatewayResultCode.GW_RES_INVALID_PARAM
 
-        return GatewayResultCode.GW_RES_OK
+        return wmm.GatewayResultCode.GW_RES_OK
 
     def write_config(self, config):
         # Should always be available
         try:
             stack_started = (self.proxy.StackStatus & 0x01) == 0
         except GLib.Error as e:
-            res = ReturnCode.error_from_dbus_exception(e.message)
+            res = ReturnCode.error_from_dbus_exception(str(e))
             self.logger.error(
                 "Cannot get Stack state. Problem in communication probably: %s",
                 res.name,
@@ -209,7 +206,7 @@ class Sink:
 
         # The write config has only one return code possible
         # so the last error code will be returned
-        res = GatewayResultCode.GW_RES_OK
+        res = wmm.GatewayResultCode.GW_RES_OK
 
         config_to_dbus_param = dict(
             [
@@ -226,7 +223,7 @@ class Sink:
         # Any following call will stop the stack
         for param in config_to_dbus_param:
             tmp = self._set_param(config, param, config_to_dbus_param[param])
-            if tmp != GatewayResultCode.GW_RES_OK:
+            if tmp != wmm.GatewayResultCode.GW_RES_OK:
                 # Update result code only if not success to avoid erasing
                 # previous error (only one return code)
                 res = tmp
@@ -243,11 +240,11 @@ class Sink:
             # App config not defined in new config
             self.logger.debug("Missing key app_config key in config: %s", config)
         except GLib.Error as e:
-            res = ReturnCode.error_from_dbus_exception(e.message)
+            res = ReturnCode.error_from_dbus_exception(str(e))
             self.logger.error("Cannot set App Config: %s", res.name)
         except OverflowError:
             # It may happens as protobuf has bigger container value
-            res = GatewayResultCode.GW_RES_INVALID_PARAM
+            res = wmm.GatewayResultCode.GW_RES_INVALID_PARAM
             self.logger.error("Invalid range value")
 
         # Set stack in state defined by new config or set it as it was
@@ -304,8 +301,8 @@ class Sink:
 
         dbus_to_gateway_satus = dict(
             [
-                (0, ScratchpadStatus.SCRATCHPAD_STATUS_SUCCESS),
-                (255, ScratchpadStatus.SCRATCHPAD_STATUS_NEW)
+                (0, wmm.ScratchpadStatus.SCRATCHPAD_STATUS_SUCCESS),
+                (255, wmm.ScratchpadStatus.SCRATCHPAD_STATUS_NEW)
                 # Anything else is ERROR
             ]
         )
@@ -318,13 +315,13 @@ class Sink:
         except KeyError:
             # Between 1 and 254 => Error
             self.logger.error("Scratchpad stored status has error: %s", status)
-            d["stored_status"] = ScratchpadStatus.SCRATCHPAD_STATUS_ERROR
+            d["stored_status"] = wmm.ScratchpadStatus.SCRATCHPAD_STATUS_ERROR
 
         dbus_to_gateway_type = dict(
             [
-                (0, ScratchpadType.SCRATCHPAD_TYPE_BLANK),
-                (1, ScratchpadType.SCRATCHPAD_TYPE_PRESENT),
-                (2, ScratchpadType.SCRATCHPAD_TYPE_PROCESS),
+                (0, wmm.ScratchpadType.SCRATCHPAD_TYPE_BLANK),
+                (1, wmm.ScratchpadType.SCRATCHPAD_TYPE_PRESENT),
+                (2, wmm.ScratchpadType.SCRATCHPAD_TYPE_PROCESS),
             ]
         )
         try:
@@ -348,10 +345,21 @@ class Sink:
 
         self._get_param(d, "firmware_area_id", "FirmwareAreaId")
 
+        try:
+            seq, crc, action, param = self.proxy.GetTargetScratchpad()
+            d["target_action"] = wmm.ScratchpadAction(
+                action + 1
+            )  # Plus one as dualmcu version starts at 0, and we start at 1
+            d["target_seq"] = seq
+            d["target_crc"] = crc
+            d["target_param"] = param
+        except GLib.Error:
+            self.logger.warning("Cannot get Target Scratchpad")
+
         return d
 
     def process_scratchpad(self):
-        ret = GatewayResultCode.GW_RES_OK
+        ret = wmm.GatewayResultCode.GW_RES_OK
         restart = False
         try:
             # Stop the stack if not already stopped
@@ -360,25 +368,25 @@ class Sink:
                 restart = True
         except GLib.Error:
             self.logger.error("Sink in invalid state")
-            return GatewayResultCode.GW_RES_INVALID_SINK_STATE
+            return wmm.GatewayResultCode.GW_RES_INVALID_SINK_STATE
 
         try:
             self.proxy.ProcessScratchpad()
         except GLib.Error as e:
-            ret = ReturnCode.error_from_dbus_exception(e.message)
+            ret = ReturnCode.error_from_dbus_exception(str(e))
             self.logger.error("Could not restore sink's state: %s", ret.name)
 
         if restart:
             try:
                 self.proxy.SetStackState(True)
             except GLib.Error as e:
-                ret = ReturnCode.error_from_dbus_exception(e.message)
+                ret = ReturnCode.error_from_dbus_exception(str(e))
                 self.logger.debug("Sink in invalid state: %s", ret.name)
 
         return ret
 
     def upload_scratchpad(self, seq, file):
-        ret = GatewayResultCode.GW_RES_OK
+        ret = wmm.GatewayResultCode.GW_RES_OK
         restart = False
         try:
             # Stop the stack if not already stopped
@@ -387,7 +395,7 @@ class Sink:
                 restart = True
         except GLib.Error:
             self.logger.error("Sink in invalid state")
-            return GatewayResultCode.GW_RES_INVALID_SINK_STATE
+            return wmm.GatewayResultCode.GW_RES_INVALID_SINK_STATE
 
         try:
             self.proxy.UploadScratchpad(seq, file)
@@ -395,11 +403,11 @@ class Sink:
                 "Scratchpad loaded with seq %d on sink %s", seq, self.sink_id
             )
         except GLib.Error as e:
-            ret = ReturnCode.error_from_dbus_exception(e.message)
+            ret = ReturnCode.error_from_dbus_exception(str(e))
             self.logger.error("Cannot upload local scratchpad: %s", ret.name)
         except OverflowError:
             # It may happens as protobuf has bigger container value
-            ret = GatewayResultCode.GW_RES_INVALID_PARAM
+            ret = wmm.GatewayResultCode.GW_RES_INVALID_PARAM
             self.logger.error("Invalid range value")
 
         if restart:
@@ -407,11 +415,68 @@ class Sink:
                 # Restart sink if we stopped it for this request
                 self.proxy.SetStackState(True)
             except GLib.Error as e:
-                ret = ReturnCode.error_from_dbus_exception(e.message)
+                ret = ReturnCode.error_from_dbus_exception(str(e))
                 self.logger.error("Could not restore sink's state: %s", ret.name)
 
         return ret
 
+    def set_target_scratchpad(self, action, target_seq, target_crc, param):
+        ret = wmm.GatewayResultCode.GW_RES_OK
+
+        if (
+            action == wmm.ScratchpadAction.ACTION_NO_OTAP
+            or action == wmm.ScratchpadAction.ACTION_LEGACY_OTAP
+        ):
+            # There is no target with those actions, default to 0
+            target_seq = 0
+            target_crc = 0
+            param = 0
+        else:
+            # check params in case there is an action with valid target
+            try:
+                if target_seq is None:
+                    # Target seq not specified, take local one
+                    target_seq = self.proxy.StoredSeq
+
+                if target_crc is None:
+                    # Target crc not specified, take local one
+                    target_crc = self.proxy.StoredCrc
+            except GLib.Error as e:
+                ret = ReturnCode.error_from_dbus_exception(str(e))
+                self.logger.error(
+                    "Cannot get local scratchpad info for set_target: %s", ret.name
+                )
+
+            if target_seq == 0:
+                self.logger.error("Seq 0 is not a valid target")
+                return wmm.GatewayResultCode.GW_RES_INVALID_PARAM
+
+        try:
+            # If there is no param for the action, default it to 0 before
+            # calling sink service through DBUS
+            if param is None:
+                param = 0
+
+            # Do a minus 1 as action are shifted by one with dual mcu
+            res = self.proxy.SetTargetScratchpad(
+                target_seq, target_crc, action.value - 1, param
+            )
+            self.logger.info(
+                "Scratchpad target set to Action %s (%d) with seq"
+                " %d and crc %d on sink %s (res=%s)",
+                action,
+                param,
+                target_seq,
+                target_crc,
+                self.sink_id,
+                res,
+            )
+        except GLib.Error as e:
+            ret = ReturnCode.error_from_dbus_exception(str(e))
+            self.logger.error("Cannot set target scratchpad: %s", ret.name)
+
+        return ret
+
 
 class SinkManager:
     "Helper class to manage the Sink list"
